// LOABoard.java
// CS3243 Homework 1: Lines of Action

// Lines of Action board. The coordinates are zero-based.

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class LOABoard {
	// Internal data structures to hold the board. For efficiency reasons,
	// we also store the number of checkers in each line. The square at the
	// ith row and jth column is considered to be in the
	// (size - 1 - i + j)th forward diagonal and the (i + j)th backward
	// diagonal. To maintain the line counts, this board implementation
	// always use the setSquare() method instead of directly assigning to
	// the board array.
	private int size;
	private int numDiags;
	private LOAColor[][] board;
	private int[] countRows;
	private int[] countCols;
	private int[] countForwardDiags;
	private int[] countBackwardDiags;

	// Constructor that creates an empty board.
	public LOABoard(int size) {
		if (size < 3)
			throw new IllegalArgumentException("Size must be at least 3.");
		this.size = size;
		this.numDiags = 2 * this.size - 1;
		this.board = new LOAColor[this.size][this.size];
		this.countRows = new int[this.size];
		this.countCols = new int[this.size];
		this.countForwardDiags = new int[this.numDiags];
		this.countBackwardDiags = new int[this.numDiags];
		for (int i = 0; i < this.size; i++)
			for (int j = 0; j < this.size; j++)
				this.board[i][j] = LOAColor.EMPTY;
		for (int i = 0; i < this.size; i++)
			this.countRows[i] = 0;
		for (int j = 0; j < this.size; j++)
			this.countCols[j] = 0;
		for (int k = 0; k < this.numDiags; k++)
			this.countForwardDiags[k] = 0;
		for (int k = 0; k < this.numDiags; k++)
			this.countBackwardDiags[k] = 0;
	}

	// Constructor that creates a board. If init is true, then we put black
	// squares on the topmost and bottommost rows, and we put white squares
	// on the leftmost and rightmost columns.
	public LOABoard(int size, boolean init) {
		this(size);
		if (init) {
			for (int j = 1; j < this.size - 1; j++) {
				setSquare(0, j, LOAColor.BLACK);
				setSquare(this.size - 1, j, LOAColor.BLACK);
			}
			for (int i = 1; i < this.size - 1; i++) {
				setSquare(i, 0, LOAColor.WHITE);
				setSquare(i, this.size - 1, LOAColor.WHITE);
			}
		}
	}

	// Constructor that creates a board from a grid string. The string is
	// assumed to contain only 'X' for black, 'O' for white, and any other
	// non-whitespace characters are assumed to be empty squares. In
	// particular, coordinates are not supported.
	public LOABoard(int size, String input) {
		this(size);
		input = input.replaceAll("\\s", "");
		if (input.length() != this.size * this.size)
			throw new IllegalArgumentException("Illegal input grid.");
		input = input.toUpperCase();
		int index = 0;
		for (int i = 0; i < this.size; i++) {
			for (int j = 0; j < this.size; j++) {
				char c = input.charAt(index);
				switch (c) {
					case 'X':
						setSquare(i, j, LOAColor.BLACK);
						break;
					case 'O':
						setSquare(i, j, LOAColor.WHITE);
						break;
					default:
						break;
				}
				index++;
			}
		}
	}

	// Copy constructor. Makes a deep copy of the internal structures.
	public LOABoard(LOABoard board) {
		this(board.getSize());
		for (int i = 0; i < this.size; i++)
			for (int j = 0; j < this.size; j++)
				setSquare(i, j, board.getSquare(i, j));
	}

	// Utility method that checks whether a row is within the legal range.
	private boolean isLegalRow(int row) {
		return (row >= 0 && row < this.size);
	}

	// Utility method that checks whether a column is within the legal
	// range.
	private boolean isLegalCol(int col) {
		return (col >= 0 && col < this.size);
	}

	// Utility method that throws an IllegalArgumentException if the row is
	// out of bounds.
	private void checkRow(int row) {
		if (!isLegalRow(row))
			throw new IllegalArgumentException("Invalid row number.");
	}

	// Utility method that throws an IllegalArgumentException if the column
	// is out of bounds.
	private void checkCol(int col) {
		if (!isLegalCol(col))
			throw new IllegalArgumentException("Invalid column number.");
	}

	// Utility method to compute the forward diagonal index.
	private int getForwardDiagIndex(int row, int col) {
		return (this.size - 1 - row + col);
	}

	// Utility method to compute the backward diagonal index.
	private int getBackwardDiagIndex(int row, int col) {
		return (row + col);
	}

	// Returns the size of the board.
	public int getSize() {
		return this.size;
	}

	// Gets the color of the specified square.
	public LOAColor getSquare(int row, int col) {
		checkRow(row);
		checkCol(col);
		return this.board[row][col];
	}

	// Sets the specified square with the specified color. Increments or
	// decrements line counters where necessary.
	public void setSquare(int row, int col, LOAColor color) {
		checkRow(row);
		checkCol(col);
		if (this.board[row][col] == LOAColor.EMPTY && color != LOAColor.EMPTY) {
			int fDiag = getForwardDiagIndex(row, col);
			int bDiag = getBackwardDiagIndex(row, col);
			this.countRows[row]++;
			this.countCols[col]++;
			this.countForwardDiags[fDiag]++;
			this.countBackwardDiags[bDiag]++;
		}
		else if (this.board[row][col] != LOAColor.EMPTY && color == LOAColor.EMPTY) {
			int fDiag = getForwardDiagIndex(row, col);
			int bDiag = getBackwardDiagIndex(row, col);
			this.countRows[row]--;
			this.countCols[col]--;
			this.countForwardDiags[fDiag]--;
			this.countBackwardDiags[bDiag]--;
		}
		this.board[row][col] = color;
	}

	// Utility method that counts the number of checkers on the line given
	// by (row, col) + lambda * (rowDir, colDir), where lambda ranges across
	// all integers. The specified (row, col) and (rowDir, colDir) is
	// assumed to be valid.
	private int getNumCheckers(int row, int col, int rowDir, int colDir) {
		if (rowDir == 0 && colDir == 0)
			throw new IllegalArgumentException("Unknown direction.");
		if (rowDir == 0)
			return this.countRows[row];
		if (colDir == 0)
			return this.countCols[col];
		if (rowDir == colDir)
			return this.countForwardDiags[getForwardDiagIndex(row, col)];
		if (rowDir == -colDir)
			return this.countBackwardDiags[getBackwardDiagIndex(row, col)];
		throw new IllegalArgumentException("Unknown direction.");
	}

	// Makes a move. Throws an IllegalArgumentException if a move is
	// invalid.
	public void move(LOAMove move) {
		if (!isLegalMove(move))
			throw new IllegalArgumentException("Illegal move.");
		int startRow = move.getStartRow();
		int startCol = move.getStartCol();
		int endRow = move.getEndRow();
		int endCol = move.getEndCol();
		LOAColor color = this.board[startRow][startCol];
		setSquare(startRow, startCol, LOAColor.EMPTY);
		setSquare(endRow, endCol, color);
	}

	// Checks whether the specified move is a legal move.
	public boolean isLegalMove(LOAMove move) {
		// Check the validity of the squares.
		int startRow = move.getStartRow();
		int startCol = move.getStartCol();
		int endRow = move.getEndRow();
		int endCol = move.getEndCol();
		if (!isLegalRow(startRow) || !isLegalCol(startCol) ||
		  !isLegalRow(endRow) || !isLegalCol(endCol))
			return false;
		// Check the validity of the direction.
		int diffRow = endRow - startRow;
		int diffCol = endCol - startCol;
		int absDiffRow = Math.abs(diffRow);
		int absDiffCol = Math.abs(diffCol);
		if ((diffRow == 0 && diffCol == 0) ||
		  (diffRow != 0 && diffCol != 0 && absDiffRow != absDiffCol))
			return false;
		int rowDir = ((diffRow == 0) ? 0 : (diffRow / absDiffRow));
		int colDir = ((diffCol == 0) ? 0 : (diffCol / absDiffCol));
		// Check the color of the start squares.
		LOAColor color = this.board[startRow][startCol];
		if (color == LOAColor.EMPTY)
			return false;
		// Check number of checkers in line.
		int count = getNumCheckers(startRow, startCol, rowDir, colDir);
		if ((absDiffRow > 0 && absDiffRow != count) ||
		  (absDiffCol > 0 && absDiffCol != count))
			return false;
		return isLegalMove(startRow, startCol, rowDir, colDir, count);
	}

	// Private helper method for finding whether a move from (row, col) to
	// (row + count * rowDir, col + count * colDir) is a legal move. It is
	// assumed that (row, col) is a valid square, and the line of the move
	// contains exactly count checkers.
	private boolean isLegalMove(int row, int col, int rowDir, int colDir, int count) {
		// Color cannot be empty.
		LOAColor color = this.board[row][col];
		if (color == LOAColor.EMPTY)
			return false;
		LOAColor opponentColor = LOAColor.getOpponentColor(color);
		// Check the line.
		for (int i = 1; i <= count; i++) {
			int r = row + i * rowDir;
			int c = col + i * colDir;
			if (!isLegalRow(r) || !isLegalCol(c))
				return false;
			if (i < count) {
				// Intervening square cannot have opposite
				// color.
				if (this.board[r][c] == opponentColor)
					return false;
			}
			else {
				// End square cannot be same color.
				if (this.board[r][c] == color)
					return false;
			}
		}
		// If we reach here, then the move is legal.
		return true;
	}

	// Obtains a list of legal moves by the specified color.
	public List<LOAMove> getLegalMoves(LOAColor color) {
		ArrayList<LOAMove> result = new ArrayList<LOAMove>();
		if (color == LOAColor.EMPTY)
			return result;
		for (int i = 0; i < this.size; i++)
			for (int j = 0; j < this.size; j++)
				if (this.board[i][j] == color)
					result.addAll(getLegalMoves(i, j));
		return result;
	}

	// Obtains a list of legal moves from a specified square. If the square
	// is empty, then there are no legal moves.
	public List<LOAMove> getLegalMoves(int row, int col) {
		checkRow(row);
		checkCol(col);
		ArrayList<LOAMove> result = new ArrayList<LOAMove>();
		if (this.board[row][col] == LOAColor.EMPTY)
			return result;
		result.addAll(getLegalMoves(row, col, 0, 1));
		result.addAll(getLegalMoves(row, col, 1, 0));
		result.addAll(getLegalMoves(row, col, 1, 1));
		result.addAll(getLegalMoves(row, col, 1, -1));
		Collections.sort(result);
		return result;
	}

	// Helper method that obtains a list of legal moves from a specified
	// non-empty square, in the line specified by
	// (row, col) + lambda * (rowDir, colDir), where lambda ranges across
	// all integers.
	private List<LOAMove> getLegalMoves(int row, int col, int rowDir, int colDir) {
		// Get the color of the specified square.
		LOAColor color = this.board[row][col];
		// Count the number of checkers in the line.
		int count = getNumCheckers(row, col, rowDir, colDir);
		// Create result list.
		ArrayList<LOAMove> result = new ArrayList<LOAMove>();
		// Check whether there is a legal move in the positive
		// direction.
		if (isLegalMove(row, col, rowDir, colDir, count))
			result.add(new LOAMove(row, col, row + count * rowDir, col + count * colDir));
		// Check whether there is a legal move in the negative
		// direction.
		if (isLegalMove(row, col, -rowDir, -colDir, count))
			result.add(new LOAMove(row, col, row - count * rowDir, col - count * colDir));
		// Return result list.
		return result;
	}

	// Checks whether the specified color has a legal move.
	public boolean hasLegalMove(LOAColor color) {
		if (color == LOAColor.EMPTY)
			return false;
		for (int i = 0; i < this.size; i++)
			for (int j = 0; j < this.size; j++)
				if (this.board[i][j] == color && hasLegalMove(i, j))
					return true;
		return false;
	}

	// Checks whether the specified square has a legal move. If the square
	// is empty, then it has no legal moves.
	public boolean hasLegalMove(int row, int col) {
		checkRow(row);
		checkCol(col);
		if (this.board[row][col] == LOAColor.EMPTY)
			return false;
		if (hasLegalMove(row, col, 0, 1))
			return true;
		if (hasLegalMove(row, col, 1, 0))
			return true;
		if (hasLegalMove(row, col, 1, 1))
			return true;
		if (hasLegalMove(row, col, 1, -1))
			return true;
		return false;
	}

	// Helper method that determines whether there is a legal move from
	// the specified non-empty square, in the line specified by 
	// (row, col) + lambda * (rowDir, colDir), where lambda ranges across
	// all integers.
	private boolean hasLegalMove(int row, int col, int rowDir, int colDir) {
		// Get the color of the specified square.
		LOAColor color = this.board[row][col];
		// Count the number of checkers in the line.
		int count = getNumCheckers(row, col, rowDir, colDir);
		// Check whether there is a legal move in the positive
		// direction.
		if (isLegalMove(row, col, rowDir, colDir, count))
			return true;
		// Check whether there is a legal move in the negative
		// direction.
		if (isLegalMove(row, col, -rowDir, -colDir, count))
			return true;
		// No legal moves.
		return false;
	}

	// Checks whether the specified color has a winning position.
	public boolean hasWinningPosition(LOAColor color) {
		// Make sure color is either black or white.
		if (color == LOAColor.EMPTY)
			throw new IllegalArgumentException("Color must be black or white.");
		// Create a marks array specifying the positions of the squares
		// with the specified color, count the number of such squares,
		// and obtain the row and column of a square with that color.
		boolean[][] marks = new boolean[this.size][this.size];
		int count = 0;
		int row = -1;
		int col = -1;
		for (int i = 0; i < this.size; i++) {
			for (int j = 0; j < this.size; j++) {
				if (this.board[i][j] == color) {
					marks[i][j] = true;
					if (count == 0) {
						row = i;
						col = j;
					}
					count++;
				}
				else
					marks[i][j] = false;
			}
		}
		// Sanity check.
		if (count == 0)
			return false;
		// Check whether the marks array contains a connected component.
		// This is done by counting the number of squares in one
		// component and see whether it equals the total number of
		// squares of that color.
		if (countComponent(marks, row, col) == count)
			return true;
		return false;
	}

	// Private helper method for counting the size of a colored component
	// containing the specified square. A side effect is that the
	// component will be unmarked in the process. This is a recursive
	// algorithm that performs a depth first traversal.
	private int countComponent(boolean[][] marks, int row, int col) {
		if (!isLegalRow(row) || !isLegalCol(col) || !marks[row][col])
			return 0;
		marks[row][col] = false;
		int result = 1;
		for (int i = row - 1; i <= row + 1; i++)
			for (int j = col - 1; j <= col + 1; j++)
				if (i != row || j != col)
					result += countComponent(marks, i, j);
		return result;
	}

	// Determines whether there is a winner for the current position. It is
	// assumed that the specified color has just moved, and the opponent has
	// not moved yet. Returns the color of the winner, or EMPTY if there is
	// no winner.
	public LOAColor getWinner(LOAColor color) {
		if (hasWinningPosition(color))
			return color;
		LOAColor opponentColor = LOAColor.getOpponentColor(color);
		if (hasWinningPosition(opponentColor))
			return opponentColor;
		if (!hasLegalMove(opponentColor))
			return color;
		return LOAColor.EMPTY;
	}

	// Determines whether two boards have the same position.
	public boolean equals(LOABoard board) {
		if (this == board)
			return true;
		if (this.size != board.getSize())
			return false;
		for (int i = 0; i < this.size; i++)
			for (int j = 0; j < this.size; j++)
				if (this.board[i][j] != board.getSquare(i, j))
					return false;
		return true;
	}

	// Determines whether two boards have the same position.
	public boolean equals(Object o) {
		if (o instanceof LOABoard)
			return equals((LOABoard)o);
		else
			return false;
	}

	// Outputs the board as a string without coordinates and without more
	// spaces.
	public String toString() {
		return toString(false, false);
	}

	// Outputs the board as a string without more spaces. Coordinates are
	// optional.
	public String toString(boolean showCoords) {
		return toString(showCoords, false);
	}

	// Outputs the board as a string. Coordinates and more spaces are
	// optional.
	public String toString(boolean showCoords, boolean showMoreSpace) {
		String lineSep = System.getProperty("line.separator");
		StringBuffer buffer = new StringBuffer();
		if (showCoords) {
			buffer.append(' ');
			if (showMoreSpace)
				buffer.append(' ');
			for (int j = 0; j < this.size; j++) {
				if (showMoreSpace && j > 0)
					buffer.append(' ');
				buffer.append((char)('A' + j));
			}
			if (showMoreSpace)
				buffer.append(' ');
			buffer.append(' ');
			buffer.append(lineSep);
		}
		for (int i = 0; i < this.size; i++) {
			if (showCoords) {
				buffer.append(i + 1);
				if (showMoreSpace)
					buffer.append(' ');
			}
			for (int j = 0; j < this.size; j++) {
				if (showMoreSpace && j > 0)
					buffer.append(' ');
				switch (this.board[i][j]) {
					case BLACK:
						buffer.append('X');
						break;
					case WHITE:
						buffer.append('O');
						break;
					case EMPTY:
						buffer.append('.');
						break;
					default:
						break;
				}
			}
			if (showCoords) {
				if (showMoreSpace)
					buffer.append(' ');
				buffer.append(i + 1);
			}
			buffer.append(lineSep);
		}
		if (showCoords) {
			buffer.append(' ');
			if (showMoreSpace)
				buffer.append(' ');
			for (int j = 0; j < this.size; j++) {
				if (showMoreSpace && j > 0)
					buffer.append(' ');
				buffer.append((char)('A' + j));
			}
			if (showMoreSpace)
				buffer.append(' ');
			buffer.append(' ');
			buffer.append(lineSep);
		}
		return buffer.toString();
	}
}

